from argparse import ArgumentParser
import random, copy, math
import numpy as np

def read_filepath(filepath, punct=True):
    tree, trees = [], []
    with open(filepath, "r") as f:
        for line in f:
            if line == "\n":
                if not punct:
                    tree = remove_punct(tree)
                trees.append(tree)
                tree = []
            elif line.startswith("#"):
                continue
            node = line.strip().split("\t")
            #if "." in node[0] or "-" in node[0]:
            #    continue
            if len(node) != 10:
                continue
            tree.append(node)
    return trees

def pos_accuracy(gold, pred):
    correct, total = 0., 0.
    for gtree, ptree in zip(gold, pred):
        for gnode, pnode in zip(gtree, ptree):
            g, p = gnode[3], pnode[3]
            if g == "_":
                continue
            if g == p:
                correct+=1
            total += 1
    return correct/total*100

def get_tag_vocab(treebank):
    vocab = set([])
    for tree in treebank:
        for node in tree:
            vocab.add(node[3])
    if "_" in vocab:
        vocab.remove("_")
    vocab = set(["ADJ",
                 "ADP",
                 "ADV",
                 "AUX",
                 "CCONJ",
                 "DET",
                 "INTJ",
                 "NOUN",
                 "NUM",
                 "PART",
                 "PRON",
                 "PROPN",
                 "PUNCT",
                 "SCONJ",
                 "SYM",
                 "VERB",
                 "X"])
    return vocab

def initialise_error_counts(vocab):
    counts = {tag: {"incorrect": 0., "total": 0.} for tag in vocab}
    return counts

def update_error_counts(error_count_dict, count_dict, weight, target_error_count, total_errors):
    generated_errors = 0
    new_dict = {tag: 0 for tag in error_count_dict.keys()}
    rate_dict = {tag: 0. for tag in error_count_dict.keys()}
    no_more = []
    for tag, error_count in error_count_dict.items():
        rate_dict[tag] = error_count / total_errors
        new_count = error_count * weight 
        if new_count > count_dict[tag]:
            new_count = count_dict[tag]
            no_more.append(tag)
        generated_errors += new_count
        new_dict[tag] = new_count
    missing = round(target_error_count - generated_errors)
    max_tries = 10
    tries = 0
    while missing > 1:
        new_dict, no_more, missing = distribute_missing_count(missing, new_dict,  rate_dict, count_dict, no_more)
        tries +=1
        if tries > max_tries:
            break
    return new_dict

def distribute_missing_count(missing, new_dict, rate_dict, count_dict, no_more):
    topped = 0
    for tag in new_dict.keys():
        if tag not in no_more:
            rate = rate_dict[tag]
            top_up = round(missing * rate)
            new_count = new_dict[tag] + top_up
            if new_count > count_dict[tag]:
                top_up = count_dict[tag] - new_count
                new_count = count_dict[tag]
                no_more.append(tag)
            topped += top_up
            new_dict[tag] = new_count
    missing -= topped
    return new_dict, no_more, missing

def get_error_rates(gold, pred, tag_vocab, desired_accuracy=80, high=False, other=False):
    error_count_dict = {tag: 0. for tag in tag_vocab}
    count_dict = {tag: 0. for tag in tag_vocab}
    total_errors = 0
    total_tokens = 0
    for gtree, ptree in zip(gold, pred):
        for gnode, pnode in zip(gtree, ptree):
            g, p = gnode[3], pnode[3]
            if g == "_":
                continue
            count_dict[g] += 1
            if g != p:
                error_count_dict[g]+=1
                total_errors += 1
            total_tokens += 1

    p_error_given_tag = {tag: 0. for tag in tag_vocab}
    p_tag_given_error = {tag: 0. for tag in tag_vocab}
    p_tag = {tag: 0. for tag in tag_vocab}
    
    
    target_p_error = (100 - desired_accuracy)/100
    target_error_count = target_p_error * total_tokens
    weight = target_error_count / total_errors
    

    new_error_count_dict = update_error_counts(error_count_dict, count_dict, weight, target_error_count, total_errors)
    if high:
        pass
    elif other:
        pass
    else:
        new_error_count_dict, total_errors = remove_actual_errors(new_error_count_dict, error_count_dict)
        count_dict, total_tokens = remove_actual_errors(count_dict, error_count_dict)
    
    p_error = total_errors/ total_tokens
    for tag, error_count in new_error_count_dict.items():
        p_tag_given_error[tag] = error_count  / total_errors
    
    for tag, count in count_dict.items():
        p_tag[tag] = count / total_tokens

    hmm = 0
    for tag in tag_vocab:
        if p_tag[tag] > 0:
            p_error_given_tag[tag] = (p_tag_given_error[tag] * p_error) / p_tag[tag]
        else:
            p_error_given_tag[tag] = 0.
    return p_error_given_tag 

def remove_actual_errors(counts, error_count_dict):
    new_counts = {tag: counts[tag] - c for tag, c in error_count_dict.items()}
    new_total = sum([v for k, v in new_counts.items()])
    return new_counts, new_total

def get_error_distros(gold, pred, tag_vocab):
    count_dict = initialise_distro_counts(tag_vocab)
    for gtree, ptree in zip(gold, pred):
        for gnode, pnode in zip(gtree, ptree):
            g, p = gnode[3], pnode[3]
            if g == "_":
                continue
            if g != p:
                count_dict[g][p] +=1
                count_dict[g]["total"] += 1
                
    distros = {tag: {} for tag in tag_vocab}
    for tag, counts in count_dict.items():
        for tag_e, c in counts.items():
            if tag_e == "total":
                continue
            try:
                proportion = c/counts["total"]
            except:
                proportion = 0.
            distros[tag][tag_e] = proportion
    return distros

def initialise_distro_counts(vocab):
    counts = {tag: {tag_e: 0. for tag_e in vocab if tag != tag_e} for tag in vocab}
    for t  in counts.keys():
        counts[t]["total"] = 0.
    return counts

def tag_with_errors(gold, pred, error_rates, error_distros, high=False, high_error_prob=None):
    errored = []
    errors = 0
    for g_tree, p_tree in zip(gold, pred):
        e_tree = []
        for g_node, p_node in zip(g_tree, p_tree):
            e_node = copy.deepcopy(g_node)
            r = random.random()
            tag = g_node[3]
            p_tag = p_node[3]
            
            if  tag == "_":
                e_tree.append(e_node)
                continue
            er = error_rates[tag]
            if not high:
                if tag != p_tag:
                    e_node[3] = p_tag
                    e_tree.append(e_node)
                    continue
                
                if r < er:
                    tag_e = get_random_error(tag, error_distros)
                    e_node[3] = tag_e
            else:
                if tag != p_tag:
                    er = error_rates[tag]
                    if r < high_error_prob:
                        #tag_e = get_random_error(tag, error_distros)
                        e_node[3] = p_tag #tag_e
            e_tree.append(e_node)
        errored.append(e_tree)
    return errored

def tag_with_errors_other(gold,  error_rates, error_distros):
    errored = []
    errors = 0
    for g_tree in gold:
        e_tree = []
        for g_node  in g_tree:
            e_node = copy.deepcopy(g_node)
            r = random.random()
            tag = g_node[3]
            if  tag == "_":
                e_tree.append(e_node)
                continue
            er = error_rates[tag]
            if r < er:
                tag_e = get_random_error(tag, error_distros)
                e_node[3] = tag_e
            e_tree.append(e_node)
        errored.append(e_tree)
    return errored

def get_random_error(tag, error_distros):
    distro = error_distros[tag]
    rd = random.random()
    tag_es = [key for key in distro.keys()]
    tag_e_id = 0
    cum_sum = 0.
    while  cum_sum < rd:
        try:
            cum_sum += distro[tag_es[tag_e_id]]
        except:
            cum_sum= 99.99
        tag_e_id += 1
    if tag_e_id > len(tag_es)-1:
        tag_e = random.choice(tag_es)
    else:
        tag_e = tag_es[tag_e_id]
    return tag_e


def save_treebank(treebank, outpath):
    with open(outpath, "w") as o:
        for tree in treebank:
            for node in tree:
                o.write("\t".join(node))
                o.write("\n")
            o.write("\n")

def check_error_crossover(gold, pred, errored, high):
    for g, p, e in zip(gold, pred, errored):
        for gn, pn, en in zip(g, p, e):
            gp = gn[3]
            pp = pn[3]
            ep = en[3]
            if high:
                if gp != ep and ep != pp:
                    print(gp, pp, ep)
                    return False
            if not high:
                if gp == ep and gp != pp:
                    print(gp, pp, ep)
                    return False
    return True

if __name__ == "__main__":
    argparser = ArgumentParser()
    argparser.add_argument("--fgold", required=True)
    argparser.add_argument("--fpred", required=True)
    argparser.add_argument("--accuracy", type=float, default=80)
    argparser.add_argument("--outpath", default=False)
    argparser.add_argument("--fother", default="")
    argparser.add_argument("--csv", action="store_true", default=False)
    args = argparser.parse_args()

    gold = read_filepath(args.fgold)
    pred = read_filepath(args.fpred)
    accuracy = pos_accuracy(gold, pred)
    if not args.csv:
        print("Accuracy (gold, pred):", format(accuracy, ".2f"))
    if args.accuracy > accuracy:
        high = True
        high_error_prob = (100 - args.accuracy) / (100 - accuracy)
    else:
        high = False
        high_error_prob = None

    if args.fother != "" and args.fother != args.fgold:
        trees = read_filepath(args.fother)
        other = True
    else:
        trees = gold
        other = False

    tag_vocab = get_tag_vocab(gold)
    error_rates = get_error_rates(gold, pred, tag_vocab, desired_accuracy=args.accuracy, high=high, other=other)
    error_distros = get_error_distros(gold, pred, tag_vocab)
    generated_accuracy = 0
    deviation = 0.05
    attempted = 0
    while np.absolute(generated_accuracy - args.accuracy) >  deviation:
        if attempted % 5 == 0 and not args.csv:
            print("Attempted: ", attempted, ", acc: ", generated_accuracy, " , dev: ", deviation)
        if not other:
            errored = tag_with_errors(trees, pred, error_rates, error_distros, high=high, high_error_prob=high_error_prob)
        else:
            errored = tag_with_errors_other(trees, error_rates, error_distros)
        generated_accuracy = pos_accuracy(trees, errored)
        if attempted > 0 and attempted % 10 == 0 and deviation < 0.5:
            deviation += 0.05
        attempted += 1
        if attempted == 500:
            generated_accuracy = -99.99
            deviation = 100000
        
    if args.outpath:
        save_treebank(errored, args.outpath)
    if not args.csv:
        print("Accuracy (gold, gen):", format(generated_accuracy, ".2f"))
    else:
        print(format(generated_accuracy, ".2f"))
        exit()
    difference_accuracy = pos_accuracy(pred, errored)
    if not other:
        print("Accuracy (pred, gen)):", format(difference_accuracy, ".2f"))
        if check_error_crossover(gold, pred, errored, high):
            print("Error generation maintained original errors.")
        else:
            print("Issue maintaining original errors from prediction file.")
 
        
